# Algorithmique et structures de données 2020-21

Contenu du cours 5 du 14.10.2020

******

##  Le tri par insertion

-   Algorithme appliqué au tableau: **4 7 6 1 2**
    -   1ère étape : 4 **7** 6 1 2
        -   trouver la position `pos` de 7 dans le sous-tableau trié à 1 élément : **4**
        -   décaler de 1 les éléments de la position `pos=1` à la fin du sous-tableau
        -   insérer l'élément 7 en position `pos=1`

    -   2ème étape : 4 7 **6** 1 2
        -   trouver la position `pos` de 6 dans le sous-tableau trié à 2 éléments : **4 7**
        -   décaler de 1 les éléments de la position `pos=1` à la fin du sous-tableau
        -   insérer l'élément 6 en position `pos=1`

    -   3ème étape : 4 6 7 **1** 2
        -   trouver la position `pos` de 1 dans le sous-tableau trié à 3 éléments : **4 6 7**
        -   décaler de 1 les éléments de la position `pos=0` à la fin du sous-tableau
        -   insérer l'élément 1 en position `pos=0`

    -   4ème étape : 1 4 6 7 **2**
        -   trouver la position `pos` de 2 dans le sous-tableau trié à 4 éléments : **1 4 6 7**
        -   décaler de 1 les éléments de la position `pos=1` à la fin du sous-tableau
        -   insérer l'élément 2 en position `pos=1`

    -   On obtient le tableau trié: **1 2 4 6 7**

## Tri à bulles

-   Algorithme appliqué au tableau **4 7 6 2 1**
    -   Principe : on parcours le tableau depuis et on permute les éléments successifs s'ils sont dans le désordre   
**4   7**   6   2   1                 
 4   **7   6**   2   1                 
 4   6   **7   2**   1                 
 4   6   2   **7   1**                 
 4   6   2   1   7                 
A la fin de cette traversée, le plus grand élément se trouve en dernière position.

    -   On applique à nouveau ce principe, mais sur le tableau allant de la 1ère à l'avant-dernière case du tableau   
 **4   6**   2   1   7                  
 4   **6   2**   1   7                  
 4   2   **6   1**   7                  
 4   2   1   6   7

    -   Et ainsi de suite ...

    -   En 4 étapes nécessitant 4, puis 3, puis 2, et finalement 1, opérations de comparaison-échange, on obtient un tableau trié.   
    Donc en 4+3+2+1 = 5**·**4/2 = 10 opérations, on a un tableau trié.

    -   Plus généralement, un tableau à N éléments se trie en N(N-1)/2 opérations avec le tri à bulles.

##  Type composé : `struct`

- **Exemple des chaînes de caractères**

    -   On peut définir un type pour les chaînes de caractères en stockant la longueur effective de la chaîne, c'est-à-dire la première occurrence du caractère `\0`
```C
        struct _chaine {
            char* str;
            int len;
        } chaine;
```
    -   Quelques implémentations de fonctions
```C
        chaine chaine_build(char* phrase) {
            chaine ch;
            ch.len = length(phrase);
            ch.str = malloc((ch.len+1)*sizeof(char));
            for (int i=0;i<ch.len;i++) {
                ch.str[i] = phrase[i];
            }
            return ch;
        }
    
        chaine deep_copy(chaine ch) {
            return chaine_build(ch.str);
        }
    
        // Dangereux, car plusieurs pointeurs
        // sur le même tableau de caractères !!!
        chaine shallow_copy(chaine ch) {
            chaine cpy;
            cpy.len = ch.len;
            cpy.str = ch.str;
            return cpy;
        }
    
        // Si plusieurs pointeurs sur même tableau de caractères,
        // alors autres pointeurs dessus ne seront pas mis à NULL
        void chaine_free(chaine* ch) {
            if (NULL != ch­>str)
                free(ch­>str);
                ch­>str = NULL;
            }
            ch­>len = -­1;
        }
```
    - **Illustration**

![Deep vs. shallow copy](./figures/fig_string_deep_shallow_copy.png)


##  Récursivité

-   Exemple de la factorielle : n ! = n·(n-1)·(n-2)·... ·3·2·1 = n·(n-1) !   
    Donc fact(n) = n·fact(n-1) (récursivité)  
    et fact(1) = 1 (condition d'arrêt)
```C
        int fact(int n) {
            if (n > 1) {
                return n*fact(n­1);
            } else {
                return 1;
            }
        }
        void main() {
            int f = fact(4);
        }
```

![Illustration](./figures/fig_recursivite_factorielle.png)

